package dvdrZ.uTwente.Skilift;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

public class FeedingCable extends Thread
{

    //@resource qLock(frac p) = (Perm(queue, p));
    //@resource preFork() = postJoin(100);
    //@resource postJoin(frac p) =Perm(placeOfMainCable,p)** Perm(cable,p) **  Perm(storage,p)  **  Perm(lastTime,p) ** Perm(simulation,p/2);
    //@resource carsOnCablePerm(frac p) = (\forall* int i ; 0 <= i && i < carsOnCable.length ; Perm(carsOnCable[i],p) );
    //@resource posLock(frac p) = (Perm(buffer, p) ** Perm(carsOnCable,p) ** carsOnCablePerm(p) );

    private int lastTime = 0;
    private final CarType type;
    private Storage storage;
    private boolean running = true;
    private final MainCable cable;
    private final int placeOfMainCable;
    private int placeOfStorage;
    private final int capacity;
    private final int length;
    private final int distanceBetweenCars;
    private final Simulation simulation;
    private int queue = 0;
    private int positionOfQueue;
    private Car buffer = null;
    private ReadWriteLock/*@ <posLock,true> @*/positionLock = new ReentrantReadWriteLock/*@<posLock> @*/();
    private ReadWriteLock/*@ <qLock,true> @*/queueLock = new ReentrantReadWriteLock/*@<qLock> @*/();
    private Car[] carsOnCable;


    /*@
      requires s != null && type != null && capacity > 0 && length > 0 && distanceBetweenCars< length/2 && cable !=null && placeOfMainCable >0;
      ensures Perm(lastTime,100) ** Perm(type,100) ** Perm(storage,100) ** Perm(running,100) ** Perm(cable,100) ** Perm(placeOfMainCable,100) ** Perm(placeOfStorage,100) ** Perm(capacity,100) ** Perm(length,100) ** Perm(distanceBetweenCars,100) ** Perm(simulation,100) ** Perm(positionOfQueue,100);
     @*/
    public FeedingCable(Simulation s, CarType type, int capacity, int length, int distanceBetweenCars, MainCable cable, int placeOfMainCable)
    {
        this.cable = cable;
        this.capacity = capacity;
        this.length = length;
        this.distanceBetweenCars = distanceBetweenCars;
        this.placeOfMainCable = placeOfMainCable;
        this.simulation = s;
        this.type = type;
    }

    /*@
      requires Perm(positionLock,50);
      ensures Perm(positionLock,50);
     @*/
    public ReadWriteLock getPositionLock()
    {

        return positionLock;

    }

    /*@
      requires Perm(carsOnCable,100) ** carsOnCablePerm(100);
      ensures Perm(carsOnCable,100) ** carsOnCablePerm(100);
     @*/
    private void addCarToArray(Car car)
    {
        boolean done = false;
        for (int i = 0; i < carsOnCable.length && !done; i++)
        {
            if (carsOnCable[i] == null)
            {
                carsOnCable[i] = car;
                done = true;
            }


        }

    }

    /*@
      requires Perm(carsOnCable,50) ** carsOnCablePerm(50);
      ensures Perm(carsOnCable,50) ** carsOnCablePerm(50);
     @*/
    private int carsSize()
    {
        int count = 0;
        for (Car car : carsOnCable)
        {
            if (car != null)
            {
                count++;
            }
        }
        return count;
    }

    /*@
    requires Perm(running,100);
    ensures Perm(running,100);
    @*/
    public void ShutDown()
    {
        running = false;
    }

    //@ensures preFork(); requires postJoin(100);
    @Override
    public void run()
    {


        while (running)
        {

            int simTime = simulation.getTime();
            int timeDif = lastTime - simTime;
            if (timeDif > 0)
            {
                moveCars(timeDif);
                lastTime = simTime;
            }
            try
            {
                Thread.sleep(250);
            }
            catch (InterruptedException e)
            {

            }
        }

    }

    /*@
      requires Perm(placeOfMainCable,100)** Perm(cable,100) **  Perm(storage,100) ** position > 0;
      ensures  Perm(placeOfMainCable,100)** Perm(cable,100) **  Perm(storage,100);
     @*/
    private void moveCars(int position)
    {

        boolean positionCanBeMoved = true;
        while (position > 0)
        {
            positionLock.readLock().lock();
            for (int i = 0; i < carsOnCable.length; i++)
            {
                boolean removed = false;

                Car car = carsOnCable[i];
                if (car != null)
                {
                    if (car.getLoad() > 0 && car.getPositionOnCable() == placeOfMainCable)
                    {

                        cable.getPositionLock().writeLock().lock();
                        if (cable.addCar(placeOfMainCable, car))
                        {

                            carsOnCable[i] = null;
                            ;
                            removed = true;
                        }
                        else
                        {
                            positionCanBeMoved = false;
                        }
                        cable.getPositionLock().writeLock().unlock();
                    }
                    if (!removed && car.getPositionOnCable() == placeOfStorage)
                    {
                        queueLock.readLock().lock();
                        if (storage.getCapacity() > storage.Count() && car.getLoad() == 0 && queue > 0)
                        {

                            storage.addCar(car);


                            carsOnCable[i] = null;
                        }
                        queueLock.readLock().unlock();
                    }
                    if (!removed && car.getPositionOnCable() == positionOfQueue)
                    {
                        queueLock.writeLock().lock();
                        if (car.getLoad() < car.getType().getCapacity() && queue > 0)
                        {

                            int placesInCart = car.getType().getCapacity() - car.getLoad();
                            if (queue > placesInCart)
                            {
                                car.setLoad(car.getLoad() + placesInCart);

                            }
                            else
                            {
                                car.setLoad(car.getLoad() + queue);
                            }

                        }
                        queueLock.writeLock().unlock();

                    }
                    if (removed)
                    {
                        i--;
                    }
                }
            }
            positionLock.readLock().unlock();
            positionLock.writeLock().lock();
            if (buffer != null)
            {
                buffer.setPositionOnCable(placeOfMainCable);
                addCarToArray(buffer);
            }
            queueLock.readLock().lock();
            if (queue > 0 && capacity > carsSize())
            {
                Car c = storage.getCar();
                c.setPositionOnCable(positionOfQueue);
                addCarToArray(c);
            }
            queueLock.readLock().unlock();

            if (positionCanBeMoved)
            {
                for (Car car : carsOnCable)
                {
                    if (car != null)
                        car.setPositionOnCable((car.getPositionOnCable() + 1) % length);

                }
            }
            positionLock.writeLock().unlock();
            position--;
        }

    }

    /*@
     requires Perm (distanceBetweenCars,50) **  incomingPosition > 0  && carPosition > 0;
     ensures Perm (distanceBetweenCars,50);
    @*/
    private boolean isBlocking(int incomingPosition, int carPosition)
    {
        boolean retval = false;


        if (incomingPosition - distanceBetweenCars < 0 || incomingPosition + distanceBetweenCars > length - 1)
        {
            retval = (incomingPosition - distanceBetweenCars) > carPosition && (incomingPosition + distanceBetweenCars < carPosition);

        }
        else
        {
            retval = (incomingPosition - distanceBetweenCars) > carPosition || (incomingPosition + distanceBetweenCars < carPosition);
        }
        return retval;
    }

    /*@ 
     requires Perm(buffer, 100) **  Perm(carsOnCable,100) ** carsOnCablePerm(100)  ** Perm (distanceBetweenCars,50);
     ensures  Perm(buffer, 100) **  Perm(carsOnCable,100) ** carsOnCablePerm(100)  ** Perm ( distanceBetweenCars,50);
    @*/
    private boolean isLegalEmptyPosition()
    {
        boolean retval = true;
        for (Car car : carsOnCable)
        {
            if (car != null)
            {
                if (isBlocking(placeOfMainCable, car.getPositionOnCable()))
                {
                    retval = false;


                }
            }
        }

        return retval;

    }

    /*@
      requires  Perm(buffer, 100) **  Perm(carsOnCable,100) ** carsOnCablePerm(100) ** car != null  && position > 0 && position < length -1;
      ensures   Perm(buffer, 100) **  Perm(carsOnCable,100) ** carsOnCablePerm(100) ** \result == isLegalEmptyPosition() && isLegalEmptyPosition() ?  buffer.equals(car) : true; 
     @*/
    public boolean AddCar(Car c)
    {
        boolean retval = isLegalEmptyPosition();

        if (retval)
        {
            buffer = c;

        }
        return retval;
    }


    /*@
      requires number > 0;
     @*/
    public void addToQueue(int number)
    {
        queueLock.writeLock().lock();
        queue += number;
        queueLock.writeLock().unlock();

    }


    /*@
      requires Perm(type,100);
      ensures Perm(type,100);
     @*/
    public CarType getType()
    {

        return type;
    }

    /*@
      requires Perm(capacity,100);
      ensures Perm(capacity,100);
     @*/
    public int getCapacity()
    {
        return capacity;
    }

    /*@
      requires Perm(distanceBetweenCars,100);
      ensures Perm(distanceBetweenCars,100);
     @*/
    public int getDistanceBetweenCars()
    {
        return distanceBetweenCars;
    }

    /*@
      requires Perm(length,100);
      ensures Perm(length,100);
     @*/
    public int getLength()
    {
        return length;
    }


}
