package dvdrZ.uTwente.Skilift;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

public class MainCable extends Thread
{
    //@resource carsOnCablePerm(frac p) = (\forall* int i ; 0 <= i && i < carsOnCable.length ; Perm(carsOnCable[i],p) ) ;
    //@resource cablesPerm(frac p) = (\forall* int i ; 0 <= i && i < cables.length ; Perm(cables[i],p) ) ;
    //@resource bufferPerm(frac p) = (\forall* int i ; 0 <= i && i < buffer.length ; Perm(buffer[i],p) ) ;
    //@resource positionOnTimePerm(frac p) = (\forall* int i ; 0 <= i && i < positionOnTime.length ; Perm(positionOnTime[i],p) ) ;

    //@resource posLock(frac p) = (Perm(buffer, p) ** Perm(carsOnCable,p) ** carsOnCablePerm(p) ** bufferPerm(p) ** positionOnTimePerm(p) );
    //@resource preFork() = postJoin(100);
    //@group resource postJoin(frac p) = Perm(lastTime,p) ** Perm(simulation,p) ** Perm(cables,100) ** cablesPerm(p);

    private final int capacity;
    private final int length;
    private final int distanceBetweenCars;
    private final Simulation simulation;
    private boolean running = true;
    private int[] positionOnTime;

    private ReadWriteLock /*@ <posLock,true> @*/positionLock = new ReentrantReadWriteLock/*@ <posLock> @*/();

    private int lastTime = 0;
    private int exit;
    private Car[] buffer;


    //The feedingcables attached to this cable
    private FeedingCable[] cables;

    private Car[] carsOnCable;

    //Creates a main cable
    /*@
      requires length > 0 &&  distanceBetweenCars < length /2 && capacity > 0 && s != null  ;
     @*/
    public MainCable(Simulation s, int capacity, int length, int distanceBetweenCars)
    {
        this.capacity = capacity;
        this.length = length;
        this.distanceBetweenCars = distanceBetweenCars;
        this.simulation = s;
        this.cables = new FeedingCable[length];
        this.buffer = new Car[length];
        this.carsOnCable = new Car[length];
        this.positionOnTime = new int[length];
    }

    /*@
      requires Perm(carsOnCable,100) ** carsOnCablePerm(100);
      ensures Perm(carsOnCable,100) ** carsOnCablePerm(100);
     @*/
    private void addCarToArray(Car car)
    {
        boolean done = false;
        for (int i = 0; i < carsOnCable.length && !done; i++)
        {
            if (carsOnCable[i] == null)
            {
                carsOnCable[i] = car;
                done = true;
            }


        }

    }


    //@ensures preFork(); requires postJoin(100);
    @Override
    public void run()
    {

        while (running)
        {

            int simTime = simulation.getTime();
            int timeDif = lastTime - simTime;
            if (timeDif > 0)
            {
                moveCars(timeDif);
                lastTime = simTime;

            }
            try
            {
                Thread.sleep(250);
            }
            catch (InterruptedException e)
            {

            }
        }


    }

    /*@
    requires Perm(carsOnCable,50) ** carsOnCablePerm(50);
    ensures Perm(carsOnCable,50) ** carsOnCablePerm(50);
    @*/
    private boolean isPositionOnTimeEmpty()
    {
        int count = 0;
        for (int i : positionOnTime)
        {
            if (i != 0)
            {
                count++;
            }
        }
        return count == 0;
    }

    /*@
      requires Perm(cables,50) ** cablesPerm(50) ** i > 0;
      ensures Perm(cables,50) ** cablesPerm(50) ** true ;
     @*/
    private void moveCars(int i)
    {


        while (i > 0)
        {
            positionLock.readLock().lock();
            for (int carID = 0; carID < carsOnCable.length; carID++)
            {
                Car car = carsOnCable[carID];
                if (car != null)
                {
                    boolean removed = false;

                    int position = car.getPositionOnCable();
                    if (position == exit)
                    {
                        car.setLoad(0);
                    }
                    if (cables[position] != null)
                    {
                        FeedingCable cable = cables[position];
                        if (cable.getType().equals(car.getType()))
                        {
                            if (cable.AddCar(car))
                            {
                                removed = true;
                                positionLock.writeLock().lock();
                                carsOnCable[carID] = null;
                                positionLock.writeLock().unlock();
                            }

                        }

                    }
                    if (removed)
                    {
                        carID--;
                    }

                }
            }
            positionLock.readLock().unlock();
            positionLock.writeLock().lock();

            for (int key = 0; key < length; key++)
            {
                if (cables[i] != null)
                {
                    Car car = buffer[key];
                    if (car != null)
                    {
                        car.setPositionOnCable(i);
                        addCarToArray(car);

                        buffer[key] = null;
                    }


                }


            }


            for (Car car : carsOnCable)
            {
                if (car != null)
                    car.setPositionOnCable((car.getPositionOnCable() + 1) % length);


            }
            positionLock.writeLock().unlock();
            i--;
        }


    }

    //Closes this thread gracefully by stopping the loop.

    /*@
      requires Perm(running,100);
      ensures Perm(running,100);
     @*/
    public void ShutDown()
    {
        running = false;
    }


    //Adds a feedingcable to this cable
    /*@
      requires Perm(cables[i],100);
      ensures Perm(cables[i],100);
     @*/
    public void addFeedingCable(int i, FeedingCable cable)
    {
        positionLock.writeLock().lock();
        cables[i] = cable;
        buffer[i] = null;
        positionLock.writeLock().unlock();
    }


    //gets the Lock placeable on the position;
    /*@
      requires Perm(positionLock,50);
      ensures Perm(positionLock,50);
     @*/
    public ReadWriteLock getPositionLock()
    {
        return positionLock;
    }


    /*@ 
      requires Perm(distanceBetweenCars, 100) ** incomingPosition != null && incomingPosition > 0 && carPosition != null && carPosition > 0;
      ensures Perm(distanceBetweenCars, 100);
     @*/
    private boolean isBlocking(int incomingPosition, int carPosition)
    {
        boolean retval = false;


        if (incomingPosition - distanceBetweenCars < 0 || incomingPosition + distanceBetweenCars > length - 1)
        {
            retval = (incomingPosition - distanceBetweenCars) > carPosition && (incomingPosition + distanceBetweenCars < carPosition);

        }
        else
        {
            retval = (incomingPosition - distanceBetweenCars) > carPosition || (incomingPosition + distanceBetweenCars < carPosition);


        }
        return retval;
    }

    /*@
      requires Perm(carsOnCable,100 ) ** Perm(buffer,100);
      ensures  Perm(carsOnCable, 100) ** Perm(buffer, 100); 
     @*/
    private boolean isLegalEmptyPosition(int position)
    {
        boolean retval = true;
        if (buffer[position] == null)
        {
            for (Car c : carsOnCable)
            {
                if (c != null)
                {
                    if (isBlocking(position, c.getPositionOnCable()))
                    {
                        retval = false;


                    }
                }
            }
        }
        else
        {
            retval = false;
        }

        return retval;
    }

    //Checks if a carriage can be added, if so, it adds the carriage, if not, it returns false, if it can be added, it adds the carriage, and returns true
    /*@
      requires Perm(buffer[position], 100) ** positionOnTimePerm(100) ** Perm(buffer[position],100) ** car != null  && position > 0 && position < length -1;
      ensures  Perm(buffer[position], 100) ** positionOnTimePerm(100) ** Perm(buffer[position],100) ** \result == isLegalEmptyPosition(position) && isLegalEmptyPosition ?  buffer.get(position).equals(car) : true ;
     @*/
    public boolean addCar(int position, Car car)
    {
        boolean retval = false;

        if (isLegalEmptyPosition(position))
        {

            if (isPositionOnTimeEmpty())
            {
                retval = true;
            }
            else if (positionOnTime[position] != 0)
            {
                int time = positionOnTime[position];
                retval = true;
                for (int i : positionOnTime)
                {

                    if (i != 0 && time > i)
                    {
                        retval = false;
                    }
                }
            }
            else
            {
                positionOnTime[position] = simulation.getTime();
            }
        }

        if (retval)
        {
            buffer[position] = car;
            positionOnTime[position] = 0;
        }

        return retval;

    }


    //gets the maximum amount of carriages on the cable
    /*@
      requires Perm(capacity,50);
      ensures Perm(capacity,50);
     @*/
    public int getCapacity()
    {
        return capacity;
    }

    //gets the minimum distance cars can have to each other
    /*@
       requires Perm(distanceBetweenCars,50);
       ensures Perm(distanceBetweenCars,50);    
     @*/
    public int getDistanceBetweenCars()
    {
        return distanceBetweenCars;
    }

    //gets the length of the cable
    /*@
      requires Perm(length,50);
      ensures Perm(length,50);
     @*/
    public int getLength()
    {
        return length;
    }

}
